
var bulk_connection;
var bulk_opened = false;
var bulk_in_ep, bulk_out_ep;

var sequence = 0x0;

const REVO_BULK_IFNUM   = 4;

const TARGET_UPGRADE 			= 0x01;
const TARGET_VERSIONRETRIEVAL 	= 0x02;
const TARGET_DOWNLOADLOG 		= 0x03;
const TARGET_CLICMD 			= 0x04;
const TARGET_CLIUPLOAD 			= 0x05;
const TARGET_CLIDOWNLOAD 		= 0x06;
const TARGET_CANCELOP 			= 0x07;
const TARGET_ORT 				= 0x08;
const TARGET_NACK 				= 0x0A;

/* Max XFER chunk size */
const MAX_BLK_XFER_SIZE = 64;

//STANDARD PACKET FORMAT:
// Multi byte entries are encoded in LittleEndian.
/*******************************************************************************
* [1-byte] [4-byte ] [1-byte] [n-byte] [ 57-n-byte ] [1-byte]
* [TARGET] [seq.No ] [ Size ] [ Data ] [ 0 padding ] [ CRC  ]
*******************************************************************************/

const CRC_DI = +0x7;
var crc_init = false;
var s_crc8_table;

function DataPacket_initCRC8Table()
{
	var i = 0x0, j = 0x0;
	var crc = 0x0;

	if (! crc_init)	{
		var crc_tab = new ArrayBuffer(256);

		s_crc8_table = new Uint8Array(crc_tab);
		for (i = 0; i < 256; i++) {
			crc = i;
			for (j = 0; j < 8; j++)
				crc = (crc << 1) ^ ((crc & 0x80) ? CRC_DI : 0);

			s_crc8_table[i] = crc & 0xFF;
			//console.log('crctab['+i+'] = ' + s_crc8_table[i]);
		}
		crc_init = true;
	}
}

function DataPacket_ComputeChecksum(buf, size) {
    var crc = +0x00;

	if (! crc_init)
		DataPacket_initCRC8Table();
    for (i = 0; i < size; i++)
		crc = s_crc8_table[(buf[i] ^ crc) & 0xff] & 0xff;
    return crc ^ 0xff;
}

function CloseBulkConnection() {
	if (bulk_opened) {
		console.log('Closing usb bulk connection');
		chrome.usb.releaseInterface(bulk_connection, bulk_interface, function() {});
		chrome.usb.closeDevice(bulk_connection);
		bulk_opened = false;
	}
	bulk_connection = null;
	bulk_interface = -1;
	chrome.runtime.getBackgroundPage(function(bkg) {
		bkg._bulk_opened = bulk_opened;
		bkg._bulk_connection = bulk_connection;
		bkg._bulk_interface = bulk_interface;
	});
}

function dumpPacket(rxBuf, len)
{	
	var i;

	for (i = 0; i < len && i < 64; i++)
		console.log('rx [' + i + '] = ' + rxBuf[i]);
}		

function RecvBulkPacketData(expectedCmd, statusUpdater, resultfunc) {
	//var gti_rd = { "direction": "in", "endpoint": bulk_in_ep, "length": 1024 * 1024 * 10, "timeout": 1000 };
	var gti_rd = { "direction": "in", "endpoint": bulk_in_ep, "length": 64, "timeout": 1000 };

	chrome.usb.bulkTransfer(bulk_connection, gti_rd, function(transinfo) {
		if (chrome.runtime.lastError && (chrome.runtime.lastError.message.search('timed out') < 0)) {
			console.log('blk rx xfer failed:' + chrome.runtime.lastError.message);
			resultfunc(false, null, 0, 0, 0);
		} else {
			var packetLen = transinfo.data.byteLength;
			var packetCount = 0;
			
			console.log('rxblk xfer len=' + packetLen + ' transres=' + transinfo.resultCode );
	
			if (transinfo.resultCode == 1 || packetLen < 64) {
				// no more data right now, so tell caller that by -1 seq
				//
				resultfunc(true, null, expectedCmd, 0, -1);
				return;
			}
			while (packetLen > 0) {
				var rxArr = transinfo.data.slice(packetCount * 64, (packetCount + 1) * 64);
				var rxBuf = new Uint8Array(rxArr);
				var i;

				var cmd = rxBuf[0];
				var seq = rxBuf[1] + (rxBuf[2] << 8) + (rxBuf[3] << 16) + (rxBuf[4] << 24);
				var len = rxBuf[5];
				var crc = rxBuf[63];

				packetCount++;
				packetLen -= 64;

				// verify cmd
				//
				if (cmd != expectedCmd) {
					console.log('Cmd ' + cmd + ' is not expected ' + expectedCmd);
					dumpPacket(rxBuf, 64);
					resultfunc(false, rxArr, cmd, len, seq);
					return;
				}
				// verify crc
				//
				var calccrc = DataPacket_ComputeChecksum(rxBuf, 63);
				
				if (calccrc != crc) {
					console.log('CRC Error, got ' + crc + ' wanted ' + calccrc);
					dumpPacket(rxBuf, 64);
					resultfunc(false, rxArr, cmd, len, seq);
					return;
				}
				resultfunc(true, rxArr, cmd, len, seq);
			}
			// signal end-of-data
			resultfunc(true, null, expectedCmd, 0, -1);
		}
	});
}

function RecvBulkPacketResponse(timeout, statusUpdater, resultfunc) {
	if (timeout < 0)
		optimeout = 2;
	else
		optimeout = timeout;
	var gti_rd = { "direction": "in", "endpoint": bulk_in_ep, "length": 64, "timeout": optimeout };

	chrome.usb.bulkTransfer(bulk_connection, gti_rd, function(transinfo) {
		if (chrome.runtime.lastError) {
			if ( (timeout < 0 && (chrome.runtime.lastError.message.search('timed out') >= 0)) ||
			  (statusUpdater == 'Ending Upgrade') ) {
				// not requiring data, and timed out, or end of upgrade
				resultfunc(true, '');
			} else {
				console.log('resp blk rx xfer ' + transinfo.data.byteLength + ' failed:' + chrome.runtime.lastError.message);
				resultfunc(false, '');
			}
		} else {
			var packetLen = transinfo.data.byteLength;
			var rxBuf = new Uint8Array(transinfo.data);
			var len = rxBuf[5] & 0xff;
			var seq = rxBuf[1] + (rxBuf[2] << 8) + (rxBuf[3] << 16) + (rxBuf[4] << 24);
			var crc = rxBuf[63];
			var i;
			
			console.log('respblk xfer ' + rxBuf[0] + ' len=' + packetLen + ' transres=' + transinfo.resultCode);

			// verify crc
			//
			var calccrc = DataPacket_ComputeChecksum(rxBuf, 63);
			
			if (calccrc != crc)
				console.log('CRC Error, got ' + crc + ' wanted ' + calccrc);

			response = '';
			for (i = 0; i < len; i++) {
				if (rxBuf[6 + i] === 0)
					break;
				response += String.fromCharCode(rxBuf[6 + i]);
			}
			switch (rxBuf[0]) {
			case TARGET_VERSIONRETRIEVAL:
				//console.log(' version resp= ' + response);
				resultfunc(true, response);
				break;
			case TARGET_DOWNLOADLOG:
				//console.log(' downloadlog resp= ' + response);
				resultfunc(true, response);
				break;
			case TARGET_UPGRADE:
				console.log(' upgrade= ' + response);
				resultfunc(true, response);
				break;
			case TARGET_NACK:
				console.log(' NACK=' + response);
				if (rxBuf[6] === 0)
					response += ' CRC Error';
				else
					response += ' Sequence Error: ' + seq;
				resultfunc(false, response);
				break;
			default:
				console.log('Unhandled response ' + rxBuf[0]);
				resultfunc(false, '');
				break;
			}
		}
	});
}

function SendBulkPacketData(getresponse, packet, statusUpdater, resultfunc) {

	if (packet.byteLength > 1024*1024)
		timeout = 5*60000; // 5 minutes!
	if (packet.byteLength > 8192)
		timeout = 60000;
	else if (packet.byteLength > 4096)
		timeout = 8000;
	else if (packet.byteLength > 1024)
		timeout = 6000;
	else
		timeout = 4000;

	var gti_wr = { "direction": "out", "endpoint": bulk_out_ep, "data": packet,	"timeout": timeout };

	chrome.usb.bulkTransfer(bulk_connection, gti_wr, function(transinfo) {
		if (chrome.runtime.lastError) {
			console.log('tx blk xfer ' + ' failed:' + chrome.runtime.lastError.message);

			resultfunc(false, '');
		}
		else {
		  if (getresponse) { // get response packet and dispatch
				  //console.log('cmd tx blk xfer ' + ' transres=' + transinfo.resultCode);
				  RecvBulkPacketResponse(60000, statusUpdater, resultfunc);
				  //RecvBulkPacketResponse(2*60000, statusUpdater, resultfunc);
			}
			else 
				  resultfunc(true, '');
		}
	});
}

function FlushBulkEndpoint(timeout, getresponse, packet, statusUpdater, resultfunc, onFlushed) {
	var gti_rd = { "direction": "in", "endpoint": bulk_in_ep, "length": 1024 * 1024, "timeout": timeout };

	onFlushed(getresponse, packet, statusUpdater, resultfunc);
  if (1) return;

	chrome.usb.bulkTransfer(bulk_connection, gti_rd, function(transinfo) {
		if (chrome.runtime.lastError && transinfo.data.byteLength < 64) {
		 console.log('flush blk rx xfer ' + transinfo.data.byteLength + ' failed:' + chrome.runtime.lastError.message);
			onFlushed(getresponse, packet, statusUpdater, resultfunc);
		} else {
			var packetLen = transinfo.data.byteLength;
			
			//console.log('flushed blk xfer len=' + packetLen + ' transres=' + transinfo.resultCode);
			FlushBulkEndpoint(timeout + 100, getresponse, packet, statusUpdater, resultfunc, onFlushed);
		}
	});
}

function OpenBulkInterface(getresponse, packet, statusUpdater, resultfunc) {
	chrome.usb.getDevices(USB_DEVICE_INFO2, function(devices) {
		if (devices) {
			if (devices.length > 0) {
				if(! bulk_opened) {
					console.log('Try to open ' + devices[0].productName);
					chrome.usb.openDevice(devices[0], function(conhandle) {
						if (chrome.runtime.lastError) {
							console.log('blk open failed:' + chrome.runtime.lastError.message);
						} else {
							console.log('dev opened');
							// enumerate interfaces and look for the one that has class 0 and
							// is past the first few and has bulk endpoints
							//
							chrome.usb.listInterfaces(conhandle, function(descriptors) {
								var d;
								// note start after 0, we know 0,1 are not what we want
								for (d = 2; d < descriptors.length; d++) {
								  console.log('USB interface number:' + descriptors[d].interfaceNumber);
									//if(descriptors[d].interfaceClass == 0 && descriptors[d].interfaceSubclass == 0) {
									if (descriptors[d].interfaceNumber == REVO_BULK_IFNUM) {
										break;
									}
								}
								if (d < descriptors.length) {

									console.log('claim interface ' + descriptors[d].interfaceNumber + ' index ' + d);
									chrome.usb.claimInterface(conhandle, descriptors[d].interfaceNumber, function() { 
										// enumerate endpoints to get in/out 
										var endpoints = descriptors[d].endpoints;
										var e;
										for (e = 0; e < endpoints.length; e++) {
											if (endpoints[e].type == 'bulk') {
												if (endpoints[e].direction == 'out')
													bulk_out_ep = endpoints[e].address;
												else if (endpoints[e].direction == 'in')
													bulk_in_ep = endpoints[e].address;
											}
										}
										console.log('claimed');
 										bulk_connection = conhandle;   
  										bulk_interface  = descriptors[d].interfaceNumber;
  										bulk_opened = true;
  										chrome.runtime.getBackgroundPage(function(bkg) {
  											bkg._bulk_opened = bulk_opened;
  											bkg._bulk_connection = bulk_connection;
  											bkg._bulk_interface = bulk_interface;
  											bkg._bulk_in_ep = bulk_in_ep;
  											bkg._bulk_out_ep = bulk_out_ep;

   											SendBulkPacket(getresponse, packet, statusUpdater, resultfunc);
	                    });
									});
								}
							});
						}
					});
				}
			} else {
				console.log('No device found');
				chrome.runtime.getBackgroundPage(function(bkg) {
					bkg._bulk_opened = false;
					bulk_opened = false;
				});
			}
		}
	});
}

function SendBulkPacket(getresponse, packet, statusUpdater, resultfunc) {
	chrome.runtime.getBackgroundPage(function(bkg) {
		bulk_opened = bkg._bulk_opened;
		if (bulk_opened) {
			bulk_connection = bkg._bulk_connection;
			bulk_interface = bkg._bulk_interface;
			bulk_in_ep = bkg._bulk_in_ep;
			bulk_out_ep = bkg._bulk_out_ep;
		}
		if (! bulk_opened) {
			// make sure we have an interface
			OpenBulkInterface(getresponse, packet, statusUpdater, resultfunc);
		} else {
			FlushBulkEndpoint(300, getresponse, packet, statusUpdater, resultfunc, SendBulkPacketData);
		}
	});
}

function DataPacket_RunCommand(cmd, data, datalen, statusUpdater, resultfunc) {

	var cmdArray = new ArrayBuffer(64);
	var cmdBuf = new Uint8Array(cmdArray);
	var i;

	var bytes = new ArrayBuffer(datalen);
	var dataBuf = new Uint8Array(bytes);
	
	for (i = 0; i < datalen; i++)
	    dataBuf[i] = data.charCodeAt(i);

	// build packet
	//
    cmdBuf[0] = cmd;
    cmdBuf[1] = sequence & 0xff;
    cmdBuf[2] = (sequence >> 8) & 0xff;
    cmdBuf[3] = (sequence >> 16) & 0xff;
    cmdBuf[4] = (sequence >> 24) & 0xff;
    cmdBuf[5] = datalen;
	for(i = 0; i < datalen; i++) {
		cmdBuf[6 + i] = dataBuf[i];
	}
	while(i < 63)
		cmdBuf[6 + i++] = 0;
    var checksum = DataPacket_ComputeChecksum(cmdBuf, 63);
    cmdBuf[63] = +checksum;
	
	// send packet on bulk endpoint, it can take a second or so to
	// open the device, so be patient (and keep it open!)
	//
	SendBulkPacket(1, cmdArray, statusUpdater, resultfunc);	
}

function DataPacket_RunRawCommand(packet, statusUpdater, resultfunc) {

	SendBulkPacket(0, packet, statusUpdater, resultfunc);	
}

